from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypedDict

from dict2xml import dict2xml  # type: ignore

from yutto.utils.time import get_time_str_by_stamp

if TYPE_CHECKING:
    from pathlib import Path


class Actor(TypedDict):
    name: str
    role: str
    thumb: str
    profile: str
    order: int


class ChapterInfoData(TypedDict):
    start: int
    end: int
    content: str


class MetaData(TypedDict):
    title: str
    show_title: str
    plot: str
    thumb: str
    premiered: int
    dateadded: int
    actor: list[Actor]
    genre: list[str]
    tag: list[str]
    source: str
    original_filename: str
    website: str
    chapter_info_data: list[ChapterInfoData]


def metadata_value_format(metadata: MetaData, metadata_format: dict[str, str]) -> dict[str, Any]:
    formatted_metadata: dict[str, Any] = {}
    for key, value in metadata.items():
        if key in metadata_format:
            assert isinstance(value, int)
            value = get_time_str_by_stamp(value, metadata_format[key])
        formatted_metadata[key] = value
    return formatted_metadata


def write_metadata(metadata: MetaData, video_path: Path, metadata_format: dict[str, str]):
    metadata_path = video_path.with_suffix(".nfo")
    custom_root = "episodedetails"  # TODO: 不同视频类型使用不同的 root name
    # 增加字段格式化内容，后续如果需要调整可以继续调整
    user_formatted_metadata = metadata_value_format(metadata, metadata_format) if metadata_format else metadata
    xml_content = dict2xml(user_formatted_metadata, wrap=custom_root, indent="  ")  # type: ignore
    with metadata_path.open("w", encoding="utf-8") as f:  # type: ignore
        f.write(xml_content)  # type: ignore


def attach_chapter_info(metadata: MetaData, chapter_info_data: list[ChapterInfoData]):
    metadata["chapter_info_data"] = chapter_info_data


# https://wklchris.github.io/blog/FFmpeg/FFmpeg.html#id26
def write_chapter_info(title: str, chapter_info_data: list[ChapterInfoData], chapter_path: Path):
    with chapter_path.open("w", encoding="utf-8") as f:
        f.write(";FFMETADATA1\n")
        f.write(f"title={title}\n")
        for chapter in chapter_info_data:
            f.write("[CHAPTER]\n")
            f.write("TIMEBASE=1/1\n")
            f.write(f"START={chapter['start']}\n")
            f.write(f"END={chapter['end']}\n")
            f.write(f"title={chapter['content']}\n")



from __future__ import annotations

import base64
import hashlib
import random
import re
import string
import time
import urllib.parse
from typing import TYPE_CHECKING, Any, TypedDict

from yutto._typing import UserInfo
from yutto.utils.asynclib import async_cache
from yutto.utils.fetcher import Fetcher, FetcherContext

if TYPE_CHECKING:
    from httpx import AsyncClient


class WbiImg(TypedDict):
    img_key: str
    sub_key: str


wbi_img_cache: WbiImg | None = None  # Simulate the LocalStorage of the browser
dm_img_str_cache: str = base64.b64encode("".join(random.choices(string.printable, k=random.randint(16, 64))).encode())[:-2].decode()  # fmt: skip
dm_cover_img_str_cache: str = base64.b64encode("".join(random.choices(string.printable, k=random.randint(32, 128))).encode())[:-2].decode()  # fmt: skip


@async_cache(lambda _: "user_info")
async def get_user_info(ctx: FetcherContext, client: AsyncClient) -> UserInfo:
    info_api = "https://api.bilibili.com/x/web-interface/nav"
    res_json = await Fetcher.fetch_json(ctx, client, info_api)
    assert res_json is not None
    res_json_data = res_json.get("data")
    return UserInfo(
        vip_status=res_json_data.get("vipStatus") == 1,  # API 返回的是 int，如果未登录就没这个值
        is_login=res_json_data.get("isLogin"),  # API 返回的是 bool
    )


async def get_wbi_img(ctx: FetcherContext, client: AsyncClient) -> WbiImg:
    global wbi_img_cache
    if wbi_img_cache is not None:
        return wbi_img_cache
    url = "https://api.bilibili.com/x/web-interface/nav"
    res_json = await Fetcher.fetch_json(ctx, client, url)
    assert res_json is not None
    wbi_img: WbiImg = {
        "img_key": _get_key_from_url(res_json["data"]["wbi_img"]["img_url"]),
        "sub_key": _get_key_from_url(res_json["data"]["wbi_img"]["sub_url"]),
    }
    wbi_img_cache = wbi_img
    return wbi_img


def _get_key_from_url(url: str) -> str:
    return url.split("/")[-1].split(".")[0]


def _get_mixin_key(string: str) -> str:
    char_indices = [
        46, 47, 18, 2, 53, 8, 23, 32, 15, 50, 10, 31, 58, 3, 45, 35, 27, 43, 5,
        49, 33, 9, 42, 19, 29, 28, 14, 39, 12, 38, 41, 13, 37, 48, 7, 16, 24, 55,
        40, 61, 26, 17, 0, 1, 60, 51, 30, 4, 22, 25, 54, 21, 56, 59, 6, 63, 57,
        62, 11, 36, 20, 34, 44, 52,
    ]  # fmt: skip
    return "".join([string[idx] for idx in char_indices[:32]])


def encode_wbi(params: dict[str, Any], wbi_img: WbiImg):
    img_key = wbi_img["img_key"]
    sub_key = wbi_img["sub_key"]
    illegal_char_remover = re.compile(r"[!'\(\)*]")

    mixin_key = _get_mixin_key(img_key + sub_key)
    time_stamp = int(time.time())
    params_with_wts = dict(params, wts=time_stamp)
    params_with_dm = {
        **params_with_wts,
        "dm_img_list": "[]",
        "dm_img_str": dm_img_str_cache,
        "dm_cover_img_str": dm_cover_img_str_cache,
    }
    url_encoded_params = urllib.parse.urlencode(
        {
            key: illegal_char_remover.sub("", str(params_with_dm[key]))
            for key in sorted(params_with_dm.keys())
        }
    )  # fmt: skip
    w_rid = hashlib.md5((url_encoded_params + mixin_key).encode()).hexdigest()
    all_params = dict(params_with_dm, w_rid=w_rid)
    return all_params



from __future__ import annotations

from typing import Any, Callable, Generic, NoReturn, Protocol, TypeVar

T = TypeVar("T")
U = TypeVar("U")


class Option(Protocol, Generic[T]):
    def __init__(self): ...

    def is_some(self) -> bool: ...

    def is_none(self) -> bool: ...

    def map(self, fn: Callable[[T], Any]) -> Option[Any]: ...

    def unwrap(self) -> T: ...

    def unwrap_or(self, default: T) -> T:
        return self.unwrap() if self.is_some() else default

    @staticmethod
    def from_optional(value: U | None) -> Option[U]:
        return Some(value) if value is not None else None_()

    def to_optional(self) -> T | None:
        return self.unwrap() if self.is_some() else None

    def __bool__(self) -> bool:
        return self.is_some()


class Some(Option[T]):
    value: T

    def __init__(self, value: T):
        self.value = value

    def is_some(self) -> bool:
        return True

    def is_none(self) -> bool:
        return False

    def map(self, fn: Callable[[T], U]) -> Option[U]:
        return Some(fn(self.value))

    def unwrap(self) -> T:
        return self.value


class None_(Option[Any]):
    def __init__(self): ...

    def is_some(self) -> bool:
        return False

    def is_none(self) -> bool:
        return True

    def map(self, fn: Callable[[Any], Any]) -> Option[Any]:
        return None_()

    def unwrap(self) -> NoReturn:
        raise ValueError("Cannot unwrap None_ object")


def map_some(fn: Callable[[T], U], value: T | None) -> U | None:
    return Option.from_optional(value).map(fn).to_optional()



from __future__ import annotations

import asyncio
import random
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Callable, TypeVar
from urllib.parse import quote, unquote

# Temporary fix for h2 stubs not found error by using `type: ignore`,
# it may be fixed in the next release. The key PR https://github.com/python-hyper/h2/pull/1289
# has been merged in the master branch
import h2.exceptions  # type: ignore
import httpx
from httpx import AsyncClient
from typing_extensions import ParamSpec

from yutto.exceptions import MaxRetryError
from yutto.utils.asynclib import async_cache
from yutto.utils.console.logger import Logger

if TYPE_CHECKING:
    from collections.abc import Coroutine, Mapping

    from yutto.utils.file_buffer import AsyncFileBuffer

RetT = TypeVar("RetT")
InputT = ParamSpec("InputT")


class MaxRetry:
    """重试装饰器，为请求方法提供一定的重试次数

    ### Args

    - max_retry (int): 额外重试次数（如重试次数为 2，则最多尝试 3 次）
    """

    def __init__(self, max_retry: int = 2):
        self.max_retry = max_retry

    def __call__(
        self, connect_once: Callable[InputT, Coroutine[Any, Any, RetT]]
    ) -> Callable[InputT, Coroutine[Any, Any, RetT]]:
        async def connect_n_times(*args: InputT.args, **kwargs: InputT.kwargs) -> RetT:
            retry = self.max_retry + 1
            while retry:
                try:
                    return await connect_once(*args, **kwargs)
                except httpx.TimeoutException:
                    Logger.warning(f"抓取超时，正在重试，剩余 {retry - 1} 次")
                except (httpx.InvalidURL, httpx.UnsupportedProtocol) as e:
                    raise e
                except httpx.HTTPError as e:
                    await asyncio.sleep(0.5)
                    error_type = e.__class__.__name__
                    Logger.warning(f"抓取失败（{error_type}），正在重试，剩余 {retry - 1} 次")
                finally:
                    retry -= 1
            raise MaxRetryError("超出最大重试次数！")

        return connect_n_times


DEFAULT_PROXY = None
DEFAULT_TRUST_ENV = True
DEFAULT_HEADERS: dict[str, str] = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36",
    "Referer": "https://www.bilibili.com",
}
DEFAULT_COOKIES = httpx.Cookies()


class FetcherContext:
    proxy: str | None
    trust_env: bool
    headers: dict[str, str]
    cookies: httpx.Cookies
    fetch_semaphore: asyncio.Semaphore | None
    download_semaphore: asyncio.Semaphore | None

    def __init__(
        self,
        *,
        proxy: str | None = DEFAULT_PROXY,
        trust_env: bool = DEFAULT_TRUST_ENV,
        headers: dict[str, str] = DEFAULT_HEADERS,
        cookies: httpx.Cookies = DEFAULT_COOKIES,
    ):
        self.proxy = proxy
        self.trust_env = trust_env
        self.headers = headers
        self.cookies = cookies
        self.fetch_semaphore = None
        self.download_semaphore = None

    def set_fetch_semaphore(self, fetch_workers: int):
        self.fetch_semaphore = asyncio.Semaphore(fetch_workers)

    def set_download_semaphore(self, download_workers: int):
        self.download_semaphore = asyncio.Semaphore(download_workers)

    def set_sessdata(self, sessdata: str):
        self.cookies = httpx.Cookies()
        # 先解码后编码是防止获取到的 SESSDATA 是已经解码后的（包含「,」）
        # 而番剧无法使用解码后的 SESSDATA
        self.cookies.set("SESSDATA", quote(unquote(sessdata)))

    def set_proxy(self, proxy: str):
        if proxy == "auto":
            self.proxy = None
            self.trust_env = True
        elif proxy == "no":
            self.proxy = None
            self.trust_env = False
        else:
            self.proxy = proxy
            self.trust_env = False

    @asynccontextmanager
    async def fetch_guard(self):
        if self.fetch_semaphore is None:
            yield
            return
        async with self.fetch_semaphore:
            yield

    @asynccontextmanager
    async def download_guard(self):
        if self.download_semaphore is None:
            yield
            return
        async with self.download_semaphore:
            yield


class Fetcher:
    @staticmethod
    @MaxRetry(2)
    async def fetch_text(
        ctx: FetcherContext,
        client: AsyncClient,
        url: str,
        *,
        params: Mapping[str, str] | None = None,
        encoding: str | None = None,  # TODO(SigureMo): Support this
    ) -> str | None:
        async with ctx.fetch_guard():
            Logger.debug(f"Fetch text: {url}")
            Logger.status.next_tick()
            resp = await client.get(url, params=params)
            if resp.status_code != httpx.codes.OK:
                return None
            return resp.text

    @staticmethod
    @MaxRetry(2)
    async def fetch_bin(
        ctx: FetcherContext,
        client: AsyncClient,
        url: str,
        *,
        params: Mapping[str, str] | None = None,
    ) -> bytes | None:
        async with ctx.fetch_guard():
            Logger.debug(f"Fetch bin: {url}")
            Logger.status.next_tick()
            resp = await client.get(url, params=params)
            if resp.status_code != httpx.codes.OK:
                return None
            return resp.read()

    @staticmethod
    @MaxRetry(2)
    async def fetch_json(
        ctx: FetcherContext,
        client: AsyncClient,
        url: str,
        *,
        params: Mapping[str, str] | None = None,
    ) -> Any | None:
        async with ctx.fetch_guard():
            Logger.debug(f"Fetch json: {url}")
            Logger.status.next_tick()
            resp = await client.get(url, params=params)
            if resp.status_code != httpx.codes.OK:
                return None
            return resp.json()

    @staticmethod
    @MaxRetry(2)
    async def get_redirected_url(ctx: FetcherContext, client: AsyncClient, url: str) -> str:
        # 关于为什么要前往重定向 url，是因为 B 站的 url 类型实在是太多了，比如有 b23.tv 的短链接
        # 为 SEO 的搜索引擎链接、甚至有的 av、BV 链接实际上是番剧页面，一一列举实在太麻烦，而且最后一种
        # 情况需要在 av、BV 解析一部分信息后才能知道是否是番剧页面，处理起来非常麻烦（bilili 就是这么做的）
        async with ctx.fetch_guard():
            resp = await client.get(url)
            redirected_url = str(resp.url)
            if redirected_url == url:
                Logger.debug(f"Get redircted url: {url}")
            else:
                Logger.debug(f"Get redircted url: {url} -> {redirected_url}")
            Logger.status.next_tick()
            return redirected_url

    @staticmethod
    @MaxRetry(2)
    async def get_size(ctx: FetcherContext, client: AsyncClient, url: str) -> int | None:
        async with ctx.fetch_guard():
            headers = client.headers.copy()
            headers["Range"] = "bytes=0-1"
            resp = await client.get(
                url,
                headers=headers,
            )
            if resp.status_code == 206:
                size = int(resp.headers["Content-Range"].split("/")[-1])
                Logger.debug(f"Get size: {url} {size}")
                return size
            else:
                return None

    @staticmethod
    @MaxRetry(2)
    # 对于相同 session，同样的页面没必要重复 touch
    @async_cache(lambda args: f"client_id={id(args.arguments['client'])}, url={args.arguments['url']}")
    async def touch_url(ctx: FetcherContext, client: AsyncClient, url: str):
        async with ctx.fetch_guard():
            Logger.debug(f"Touch url: {url}")
            await client.get(url)

    @staticmethod
    async def download_file_with_offset(
        ctx: FetcherContext,
        client: AsyncClient,
        url: str,
        mirrors: list[str],
        file_buffer: AsyncFileBuffer,
        offset: int,
        size: int | None,
    ) -> None:
        async with ctx.download_guard():
            Logger.debug(f"Start download (offset {offset}, number of mirrors {len(mirrors)}) {url}")
            done = False
            headers = client.headers.copy()
            url_pool = [url] + mirrors
            block_offset = 0
            while not done:
                try:
                    url = random.choice(url_pool)
                    headers["Range"] = "bytes={}-{}".format(
                        offset + block_offset, offset + size - 1 if size is not None else ""
                    )
                    async with client.stream(
                        "GET",
                        url,
                        headers=headers,
                        timeout=httpx.Timeout(7, connect=3),
                    ) as resp:
                        # 如果直接用 1KiB 的话，会产生大量的块，需要消耗大量的 CPU 资源来维持顺序，
                        # 而使用 1MiB 以上或者不使用流式下载方式时，由于分块太大，
                        # 导致进度条显示的实时速度并不准，波动太大，用户体验不佳，
                        # 因此取两者折中
                        async for chunk in resp.aiter_bytes(2**16):
                            await file_buffer.write(chunk, offset + block_offset)
                            block_offset += len(chunk)
                    # TODO: 是否需要校验总大小
                    done = True

                except httpx.TimeoutException:
                    Logger.warning(f"文件 {file_buffer.file_path} 下载超时，尝试重新连接...")
                    Logger.debug(f"超时链接：{url}")
                except (httpx.HTTPError, h2.exceptions.H2Error) as e:
                    await asyncio.sleep(0.5)
                    error_type = e.__class__.__name__
                    Logger.warning(f"文件 {file_buffer.file_path} 下载出错（{error_type}），尝试重新连接...")
                    Logger.debug(f"超时链接：{url}")
                except ValueError as e:
                    # 由于 httpx 经常出现此问题，暂时捕获该问题
                    if "semaphore released too many times" not in str(e):
                        raise e
                    Logger.warning(f"文件 {file_buffer.file_path} 下载出错（{e}），尝试重新连接...")


def create_client(
    headers: dict[str, str] = DEFAULT_HEADERS,
    cookies: httpx.Cookies = DEFAULT_COOKIES,
    trust_env: bool = DEFAULT_TRUST_ENV,
    proxy: str | None = DEFAULT_PROXY,
    timeout: int | httpx.Timeout = 5,
) -> AsyncClient:
    client = httpx.AsyncClient(
        headers=headers,
        cookies=cookies,
        trust_env=trust_env,
        proxy=proxy,
        timeout=timeout,
        follow_redirects=True,
        http2=True,
        verify=False,
    )
    return client



from __future__ import annotations

import asyncio
import os
import re
import sys
from typing import TYPE_CHECKING

import biliass

from yutto.api.user_info import get_user_info
from yutto.bilibili_typing.codec import VideoCodec, audio_codec_priority_default, video_codec_priority_default
from yutto.exceptions import ErrorCode
from yutto.processor.selector import validate_episodes_selection
from yutto.utils.asynclib import initial_async_policy
from yutto.utils.console.colorful import set_no_color
from yutto.utils.console.logger import Badge, Logger, set_logger_debug
from yutto.utils.fetcher import FetcherContext, create_client
from yutto.utils.ffmpeg import FFmpeg
from yutto.utils.filter import Filter

if TYPE_CHECKING:
    import argparse

    from yutto._typing import UserInfo


def initial_validation(ctx: FetcherContext, args: argparse.Namespace):
    """初始化检查，仅执行一次"""

    if not args.no_progress:
        Logger.enable_statusbar()

    # 在使用 --no-color 或者环境变量 NO_COLOR 非空时都应该不显示颜色
    # See also: https://no-color.org/
    if args.no_color or os.environ.get("NO_COLOR"):
        set_no_color()

    # debug 设置
    if args.debug:
        set_logger_debug()
        biliass.enable_tracing()

    # 初始化异步策略，消除平台差异
    initial_async_policy()

    # proxy 校验
    if args.proxy not in ["no", "auto"] and not re.match(r"https?://", args.proxy):
        Logger.error(f"proxy 参数值（{args.proxy}）错误啦！")
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)
    ctx.set_proxy(args.proxy)

    # 大会员身份校验
    if not args.sessdata:
        Logger.info("未提供 SESSDATA，无法下载高清视频、字幕等资源哦～")
    else:
        ctx.set_sessdata(args.sessdata)
        if asyncio.run(validate_user_info(ctx, {"vip_status": True, "is_login": True})):
            Logger.custom("成功以大会员身份登录～", badge=Badge("大会员", fore="white", back="magenta", style=["bold"]))
        else:
            Logger.warning("以非大会员身份登录，注意无法下载会员专享剧集喔～")

    # 批量下载时的过滤器设置
    if args.batch_filter_start_time:
        Filter.set_timer("batch_filter_start_time", args.batch_filter_start_time)
    if args.batch_filter_end_time:
        Filter.set_timer("batch_filter_end_time", args.batch_filter_end_time)


def validate_basic_arguments(args: argparse.Namespace):
    """检查 argparse 无法检查的选项，并设置某些全局的状态"""

    ffmpeg = FFmpeg()

    download_vcodec_priority: list[VideoCodec] = video_codec_priority_default
    if args.download_vcodec_priority is not None:
        user_download_vcodec_priority = args.download_vcodec_priority
        if not user_download_vcodec_priority:
            Logger.error("download_vcodec_priority 参数值为空哦")
            sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)
        for vcodec in user_download_vcodec_priority:
            if vcodec not in video_codec_priority_default:
                Logger.error(
                    "download_vcodec_priority 参数值（{}）不满足要求哦（允许值：{{{}}}）".format(
                        vcodec, ", ".join(video_codec_priority_default)
                    )
                )
                sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)
        download_vcodec_priority = user_download_vcodec_priority
        if len(download_vcodec_priority) < len(video_codec_priority_default):
            Logger.warning(
                "download_vcodec_priority（{}）不包含所有下载视频编码（{}），不包含部分将永远不会选择哦".format(
                    ", ".join(args.download_vcodec_priority), ", ".join(video_codec_priority_default)
                )
            )

    # vcodec 检查
    vcodec_split = args.vcodec.split(":")
    if len(vcodec_split) != 2:
        Logger.error(f"vcodec 参数值（{args.vcodec}）不满足要求哦（并非使用 : 分隔的值）")
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)
    video_download_codec, video_save_codec = vcodec_split
    if video_download_codec not in download_vcodec_priority:
        Logger.error(
            "download_vcodec 参数值（{}）不满足要求哦（允许值：{{{}}}）".format(
                video_download_codec, ", ".join(download_vcodec_priority)
            )
        )
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)
    if args.download_vcodec_priority is not None and download_vcodec_priority[0] != video_download_codec:
        Logger.warning(
            f"download_vcodec 参数值（{video_download_codec}）不是优先级最高的编码（{download_vcodec_priority[0]}），可能会导致下载失败哦"
        )
    if video_save_codec not in ffmpeg.video_encodecs + ["copy"]:
        Logger.error(
            "save_vcodec 参数值（{}）不满足要求哦（允许值：{{{}}}）".format(
                video_save_codec, ", ".join(ffmpeg.video_encodecs + ["copy"])
            )
        )
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)

    # acodec 检查
    acodec_split = args.acodec.split(":")
    if len(acodec_split) != 2:
        Logger.error(f"acodec 参数值（{args.acodec}）不满足要求哦（并非使用 : 分隔的值）")
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)
    audio_download_codec, audio_save_codec = acodec_split
    if audio_download_codec not in audio_codec_priority_default:
        Logger.error(
            "download_acodec 参数值（{}）不满足要求哦（允许值：{{{}}}）".format(
                audio_download_codec, ", ".join(audio_codec_priority_default)
            )
        )
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)
    if audio_save_codec not in ffmpeg.audio_encodecs + ["copy"]:
        Logger.error(
            "save_acodec 参数值（{}）不满足要求哦（允许值：{{{}}}）".format(
                audio_save_codec, ", ".join(ffmpeg.audio_encodecs + ["copy"])
            )
        )
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)

    # cover 检查
    if not args.require_cover and args.save_cover:
        Logger.warning("没有下载封面的情况下是无法保留封面的哦～")
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)


def validate_batch_arguments(args: argparse.Namespace):
    """检查批量下载相关选项"""
    # 检查 episodes 格式（简单的正则检查，后续过滤剧集时还有完整检查）
    if not validate_episodes_selection(args.episodes):
        # TODO: 错误信息链接到相应文档，当然需要先写文档……
        Logger.error(f"选集参数（{args.episodes}）格式不正确呀～重新检查一下下～")
        sys.exit(ErrorCode.WRONG_ARGUMENT_ERROR.value)


async def validate_user_info(ctx: FetcherContext, check_option: UserInfo) -> bool:
    """UserInfo 结构和用户输入是匹配的，如果要校验则置 True 即可，估计不会有要校验为 False 的情况吧~~"""
    async with create_client(
        cookies=ctx.cookies,
        trust_env=ctx.trust_env,
        proxy=ctx.proxy,
    ) as client:
        if check_option["is_login"] or check_option["vip_status"]:
            # 需要校验
            # 这么写 if 是为了少一个 get_user_info 请求
            user_info = await get_user_info(ctx, client)
            if check_option["is_login"] and not user_info["is_login"]:
                return False
            if check_option["vip_status"] and not user_info["vip_status"]:
                return False
        return True



from __future__ import annotations

import json
from pathlib import Path

from yutto.cli.settings import YuttoSettings


def main():
    schema = YuttoSettings.model_json_schema()
    with Path("schemas/config.json").open("w") as f:
        json.dump(schema, f, indent=2)


if __name__ == "__main__":
    main()



from __future__ import annotations

import time

TIME_FULL_FMT = "%Y-%m-%d %H:%M:%S"
TIME_DATE_FMT = "%Y-%m-%d"


def get_time_stamp_by_now() -> int:
    return int(time.time())


def get_time_str_by_now(fmt: str = TIME_FULL_FMT):
    time_stamp_now = time.time()
    return get_time_str_by_stamp(time_stamp_now, fmt)


def get_time_str_by_stamp(stamp: float, fmt: str = TIME_FULL_FMT):
    local_time = time.localtime(stamp)
    return time.strftime(fmt, local_time)


def get_time_struct_by_stamp(stamp: float):
    return time.localtime(stamp)



from __future__ import annotations

import pytest

from yutto.processor.path_resolver import create_unique_path_resolver


@pytest.mark.processor
def test_unique_path():
    unique_path = create_unique_path_resolver()
    assert unique_path("a") == "a"
    assert unique_path("a") == "a (1)"
    assert unique_path("a") == "a (2)"

    assert unique_path("/xxx/yyy/zzz.ext") == "/xxx/yyy/zzz.ext"
    assert unique_path("/xxx/yyy/zzz.ext") == "/xxx/yyy/zzz (1).ext"
    assert unique_path("/xxx/yyy/zzz.ext") == "/xxx/yyy/zzz (2).ext"



from __future__ import annotations

import asyncio
import inspect
import platform
import time
from functools import wraps
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from typing_extensions import ParamSpec

from yutto.utils.console.logger import Logger

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine, Generator, Iterable

RetT = TypeVar("RetT")
P = ParamSpec("P")


def initial_async_policy():
    if platform.system() == "Windows":
        Logger.debug("Windows 平台，单独设置 EventLoopPolicy")
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore


class CoroutineWrapper(Generic[RetT]):
    coro: Coroutine[Any, Any, RetT]

    def __init__(self, coro: Coroutine[Any, Any, RetT]):
        self.coro = coro

    def __await__(self) -> Generator[Any, None, RetT]:
        return (yield from self.coro.__await__())

    def __del__(self):
        self.coro.close()


async def sleep_with_status_bar_refresh(seconds: float):
    current_time = start_time = time.time()
    while current_time - start_time < seconds:
        Logger.status.next_tick()
        await asyncio.sleep(min(1, seconds - (current_time - start_time)))
        current_time = time.time()


def async_cache(
    args_to_cache_key: Callable[[inspect.BoundArguments], str],
) -> Callable[[Callable[P, Coroutine[Any, Any, RetT]]], Callable[P, Coroutine[Any, Any, RetT]]]:
    CACHE: dict[str, RetT] = {}

    def decorator(fn: Callable[P, Coroutine[Any, Any, RetT]]) -> Callable[P, Coroutine[Any, Any, RetT]]:
        @wraps(fn)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> RetT:
            sig = inspect.signature(fn)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            cache_key = args_to_cache_key(bound_args)
            if cache_key in CACHE:
                Logger.debug(f"{fn.__name__} cache hit: {cache_key}")
                return CACHE[cache_key]
            Logger.debug(f"{fn.__name__} cache miss: {cache_key}, all cache keys: {list(CACHE.keys())}")
            return CACHE.setdefault(cache_key, await fn(*args, **kwargs))

        return wrapper

    return decorator


async def first_successful(coros: Iterable[Coroutine[Any, Any, RetT]]) -> list[RetT]:
    tasks = [asyncio.create_task(coro) for coro in coros]

    results: list[RetT] = []
    while not results:
        done, tasks = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        results = [task.result() for task in done if task.exception() is None]
    for task in tasks:
        task.cancel()
    return results


async def first_successful_with_check(coros: Iterable[Coroutine[Any, Any, RetT]]) -> RetT:
    results = await first_successful(coros)
    if not results:
        raise Exception("All coroutines failed")
    if len(set(results)) != 1:
        raise Exception("Multiple coroutines returned different results")
    return results[0]



from __future__ import annotations

from typing import TYPE_CHECKING

from yutto._typing import AvId, EpisodeData, EpisodeId, format_ids
from yutto.api.bangumi import (
    BangumiListItem,
    get_bangumi_playurl,
    get_bangumi_subtitles,
)
from yutto.api.cheese import CheeseListItem, get_cheese_playurl, get_cheese_subtitles
from yutto.api.danmaku import get_danmaku
from yutto.api.ugc_video import (
    UgcVideoListItem,
    get_ugc_video_chapters,
    get_ugc_video_playurl,
    get_ugc_video_subtitles,
)
from yutto.exceptions import (
    HttpStatusError,
    NoAccessPermissionError,
    NotFoundError,
    UnSupportedTypeError,
)
from yutto.processor.path_resolver import (
    UNKNOWN,
    PathTemplateVariableDict,
    resolve_path_template,
)
from yutto.utils.console.logger import Logger
from yutto.utils.danmaku import EmptyDanmakuData
from yutto.utils.fetcher import Fetcher, FetcherContext
from yutto.utils.metadata import attach_chapter_info

if TYPE_CHECKING:
    import argparse
    from pathlib import Path

    import httpx


async def extract_bangumi_data(
    ctx: FetcherContext,
    client: httpx.AsyncClient,
    bangumi_info: BangumiListItem,
    args: argparse.Namespace,
    subpath_variables: PathTemplateVariableDict,
    auto_subpath_template: str = "{name}",
) -> EpisodeData | None:
    try:
        avid = bangumi_info["avid"]
        cid = bangumi_info["cid"]
        name = bangumi_info["name"]
        id = bangumi_info["id"]
        if bangumi_info["is_preview"]:
            Logger.warning(f"视频（{format_ids(avid, cid)}）是预览视频（疑似未登录或非大会员用户）")
        videos, audios = (
            await get_bangumi_playurl(ctx, client, avid, cid) if args.require_video or args.require_audio else ([], [])
        )
        subtitles = await get_bangumi_subtitles(ctx, client, avid, cid) if args.require_subtitle else []
        danmaku = (
            await get_danmaku(ctx, client, cid, avid, args.danmaku_format) if args.require_danmaku else EmptyDanmakuData
        )
        metadata = bangumi_info["metadata"] if args.require_metadata else None
        cover_data = (
            await Fetcher.fetch_bin(ctx, client, bangumi_info["metadata"]["thumb"]) if args.require_cover else None
        )
        subpath_variables_base: PathTemplateVariableDict = {
            "id": id,
            "aid": str(avid.as_aid()),
            "bvid": str(avid.as_bvid()),
            "name": name,
            "title": UNKNOWN,
            "username": UNKNOWN,
            "series_title": UNKNOWN,
            "pubdate": UNKNOWN,
            "download_date": bangumi_info["metadata"]["dateadded"],
            "owner_uid": UNKNOWN,
        }
        subpath_variables_base.update(subpath_variables)
        subpath = resolve_path_template(args.subpath_template, auto_subpath_template, subpath_variables_base)
        file_path: Path = args.dir / subpath
        output_dir, filename = file_path.parent, file_path.name
        return EpisodeData(
            videos=videos,
            audios=audios,
            subtitles=subtitles,
            metadata=metadata,
            danmaku=danmaku,
            cover_data=cover_data,
            chapter_info_data=[],
            output_dir=output_dir,
            tmp_dir=args.tmp_dir or output_dir,
            filename=filename,
        )
    except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
        Logger.error(e.message)
        return None


async def extract_cheese_data(
    ctx: FetcherContext,
    client: httpx.AsyncClient,
    episode_id: EpisodeId,
    cheese_info: CheeseListItem,
    args: argparse.Namespace,
    subpath_variables: PathTemplateVariableDict,
    auto_subpath_template: str = "{name}",
) -> EpisodeData | None:
    try:
        avid = cheese_info["avid"]
        cid = cheese_info["cid"]
        name = cheese_info["name"]
        id = cheese_info["id"]
        videos, audios = (
            await get_cheese_playurl(ctx, client, avid, episode_id, cid)
            if args.require_video or args.require_audio
            else ([], [])
        )
        subtitles = await get_cheese_subtitles(ctx, client, avid, cid) if args.require_subtitle else []
        danmaku = (
            await get_danmaku(ctx, client, cid, avid, args.danmaku_format) if args.require_danmaku else EmptyDanmakuData
        )
        metadata = cheese_info["metadata"] if args.require_metadata else None
        cover_data = (
            await Fetcher.fetch_bin(ctx, client, cheese_info["metadata"]["thumb"]) if args.require_cover else None
        )
        subpath_variables_base: PathTemplateVariableDict = {
            "id": id,
            "aid": str(avid.as_aid()),
            "bvid": str(avid.as_bvid()),
            "name": name,
            "title": UNKNOWN,
            "username": UNKNOWN,
            "series_title": UNKNOWN,
            "pubdate": UNKNOWN,
            "download_date": UNKNOWN,
            "owner_uid": UNKNOWN,
        }
        subpath_variables_base.update(subpath_variables)
        subpath = resolve_path_template(args.subpath_template, auto_subpath_template, subpath_variables_base)
        file_path: Path = args.dir / subpath
        output_dir, filename = file_path.parent, file_path.name
        return EpisodeData(
            videos=videos,
            audios=audios,
            subtitles=subtitles,
            metadata=metadata,
            danmaku=danmaku,
            cover_data=cover_data,
            chapter_info_data=[],
            output_dir=output_dir,
            tmp_dir=args.tmp_dir or output_dir,
            filename=filename,
        )
    except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
        Logger.error(e.message)
        return None


async def extract_ugc_video_data(
    ctx: FetcherContext,
    client: httpx.AsyncClient,
    avid: AvId,
    ugc_video_info: UgcVideoListItem,
    args: argparse.Namespace,
    subpath_variables: PathTemplateVariableDict,
    auto_subpath_template: str = "{title}",
) -> EpisodeData | None:
    try:
        cid = ugc_video_info["cid"]
        name = ugc_video_info["name"]
        id = ugc_video_info["id"]
        videos, audios = (
            await get_ugc_video_playurl(ctx, client, avid, cid)
            if args.require_video or args.require_audio
            else ([], [])
        )
        subtitles = await get_ugc_video_subtitles(ctx, client, avid, cid) if args.require_subtitle else []
        chapter_info_data = await get_ugc_video_chapters(ctx, client, avid, cid) if args.require_chapter_info else []
        danmaku = (
            await get_danmaku(ctx, client, cid, avid, args.danmaku_format) if args.require_danmaku else EmptyDanmakuData
        )
        metadata = ugc_video_info["metadata"] if args.require_metadata else None
        if metadata and chapter_info_data:
            attach_chapter_info(metadata, chapter_info_data)
        cover_data = (
            await Fetcher.fetch_bin(ctx, client, ugc_video_info["metadata"]["thumb"]) if args.require_cover else None
        )
        owner_uid: str = (
            ugc_video_info["metadata"]["actor"][0]["profile"].split("/")[-1]
            if ugc_video_info["metadata"]["actor"]
            else UNKNOWN
        )
        username: str = (
            ugc_video_info["metadata"]["actor"][0]["name"] if ugc_video_info["metadata"]["actor"] else UNKNOWN
        )
        subpath_variables_base: PathTemplateVariableDict = {
            "id": id,
            "aid": str(avid.as_aid()),
            "bvid": str(avid.as_bvid()),
            "name": name,
            "title": UNKNOWN,
            "username": username,
            "series_title": UNKNOWN,
            "pubdate": UNKNOWN,
            "download_date": ugc_video_info["metadata"]["dateadded"],
            "owner_uid": owner_uid,
        }
        subpath_variables_base.update(subpath_variables)
        subpath = resolve_path_template(args.subpath_template, auto_subpath_template, subpath_variables_base)
        file_path: Path = args.dir / subpath
        output_dir, filename = file_path.parent, file_path.name
        return EpisodeData(
            videos=videos,
            audios=audios,
            subtitles=subtitles,
            metadata=metadata,
            danmaku=danmaku,
            cover_data=cover_data,
            chapter_info_data=chapter_info_data,
            output_dir=output_dir,
            tmp_dir=args.tmp_dir or output_dir,
            filename=filename,
        )
    except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
        Logger.error(e.message)
        return None



from __future__ import annotations

import os
import platform
import sys
from pathlib import Path
from typing import Annotated, Any, Literal, Optional

from pydantic import BaseModel, Field

from yutto.bilibili_typing.quality import (
    AudioQuality,
    VideoQuality,
)
from yutto.utils.console.logger import Logger
from yutto.utils.time import TIME_DATE_FMT

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib  # type: ignore


def xdg_config_home() -> Path:
    if (env := os.environ.get("XDG_CONFIG_HOME")) and (path := Path(env)).is_absolute():
        return path
    home = Path.home()
    if platform.system() == "Windows":
        return home / "AppData" / "Roaming"
    return home / ".config"


class YuttoBasicSettings(BaseModel):
    num_workers: Annotated[int, Field(8, gt=0)]
    video_quality: Annotated[VideoQuality, Field(127)]
    audio_quality: Annotated[AudioQuality, Field(30251)]
    vcodec: Annotated[str, Field("avc:copy")]
    acodec: Annotated[str, Field("mp4a:copy")]
    download_vcodec_priority: Annotated[Optional[list[str]], Field(None)]  # noqa: UP007
    output_format: Annotated[Literal["infer", "mp4", "mkv", "mov"], Field("infer")]
    output_format_audio_only: Annotated[
        Literal["infer", "m4a", "aac", "mp3", "flac", "mp4", "mkv", "mov"], Field("infer")
    ]
    danmaku_format: Annotated[Literal["xml", "ass", "protobuf"], Field("ass")]
    block_size: Annotated[float, Field(0.5)]
    overwrite: Annotated[bool, Field(False)]
    proxy: Annotated[str, Field("auto")]
    dir: Annotated[str, Field("./")]
    tmp_dir: Annotated[Optional[str], Field(None)]  # noqa: UP007
    sessdata: Annotated[str, Field("")]
    subpath_template: Annotated[str, Field("{auto}")]
    aliases: Annotated[dict[str, str], Field({})]
    metadata_format_premiered: Annotated[str, Field(TIME_DATE_FMT)]
    download_interval: Annotated[int, Field(0)]
    banned_mirrors_pattern: Annotated[Optional[str], Field(None)]  # noqa: UP007
    no_color: Annotated[bool, Field(False)]
    no_progress: Annotated[bool, Field(False)]
    debug: Annotated[bool, Field(False)]
    vip_strict: Annotated[bool, Field(False)]
    login_strict: Annotated[bool, Field(False)]


class YuttoResourceSettings(BaseModel):
    require_video: Annotated[bool, Field(True)]
    require_audio: Annotated[bool, Field(True)]
    require_subtitle: Annotated[bool, Field(True)]
    require_metadata: Annotated[bool, Field(False)]
    require_danmaku: Annotated[bool, Field(True)]
    require_cover: Annotated[bool, Field(True)]
    require_chapter_info: Annotated[bool, Field(True)]
    save_cover: Annotated[bool, Field(False)]


class YuttoDanmakuSettings(BaseModel):
    font_size: Annotated[Optional[int], Field(None)]  # noqa: UP007
    font: Annotated[str, Field("SimHei")]
    opacity: Annotated[float, Field(0.8)]
    display_region_ratio: Annotated[float, Field(1.0)]
    speed: Annotated[float, Field(1.0)]
    block_top: Annotated[bool, Field(False)]
    block_bottom: Annotated[bool, Field(False)]
    block_scroll: Annotated[bool, Field(False)]
    block_reverse: Annotated[bool, Field(False)]
    block_fixed: Annotated[bool, Field(False)]
    block_special: Annotated[bool, Field(False)]
    block_colorful: Annotated[bool, Field(False)]
    block_keyword_patterns: Annotated[list[str], Field([])]


class YuttoBatchSettings(BaseModel):
    with_section: Annotated[bool, Field(False)]
    batch_filter_start_time: Annotated[Optional[str], Field(None)]  # noqa: UP007
    batch_filter_end_time: Annotated[Optional[str], Field(None)]  # noqa: UP007


class YuttoSettings(BaseModel):
    basic: Annotated[YuttoBasicSettings, Field(YuttoBasicSettings())]  # pyright: ignore[reportCallIssue]
    resource: Annotated[YuttoResourceSettings, Field(YuttoResourceSettings())]  # pyright: ignore[reportCallIssue]
    danmaku: Annotated[YuttoDanmakuSettings, Field(YuttoDanmakuSettings())]  # pyright: ignore[reportCallIssue]
    batch: Annotated[YuttoBatchSettings, Field(YuttoBatchSettings())]  # pyright: ignore[reportCallIssue]


def search_for_settings_file() -> Path | None:
    settings_file = Path("yutto.toml")
    # 此时还没有设置 debug，所以 Logger.debug 永远不会输出
    if not settings_file.exists():
        Logger.debug("Settings file not found in current directory.")
        settings_file = xdg_config_home() / "yutto" / "yutto.toml"
    if not settings_file.exists():
        Logger.debug(f"Settings file not found in XDG_CONFIG_HOME ({settings_file}).")
        return None
    Logger.debug(f"Settings file found at {settings_file}.")
    return settings_file


def load_settings_file(settings_file: Path) -> YuttoSettings:
    with settings_file.open("r") as f:
        settings_raw: Any = tomllib.loads(f.read())  # pyright: ignore[reportUnknownMemberType]
    return YuttoSettings.model_validate(settings_raw)


if __name__ == "__main__":
    settings_file = search_for_settings_file()
    assert settings_file is not None
    settings = load_settings_file(settings_file)
    print(settings.model_dump())



from __future__ import annotations

import argparse
from typing import TYPE_CHECKING, Any, Literal

from yutto.__version__ import VERSION as yutto_version
from yutto.bilibili_typing.quality import (
    audio_quality_priority_default,
    video_quality_priority_default,
)
from yutto.cli.settings import YuttoSettings, load_settings_file, search_for_settings_file
from yutto.processor.parser import alias_parser, path_from_cli
from yutto.utils.console.logger import Logger
from yutto.utils.funcutils.option import map_some

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    from typing_extensions import TypeAlias


DownloadResourceType: TypeAlias = Literal["video", "audio", "subtitle", "metadata", "danmaku", "cover", "chapter_info"]
DOWNLOAD_RESOURCE_TYPES: list[DownloadResourceType] = [
    "video",
    "audio",
    "subtitle",
    "metadata",
    "danmaku",
    "cover",
    "chapter_info",
]


def parse_config_path() -> Path | None:
    pre_parser = argparse.ArgumentParser(description="yutto pre parser", add_help=False)
    pre_parser.add_argument("--config", type=path_from_cli, default=search_for_settings_file(), help="配置文件路径")
    args, _ = pre_parser.parse_known_args()
    return args.config


def cli() -> argparse.ArgumentParser:
    settings_file = parse_config_path()
    if settings_file is None:
        settings = YuttoSettings()  # pyright: ignore[reportCallIssue]
    else:
        Logger.info(f"发现配置文件 {settings_file}，加载中……")
        settings = load_settings_file(settings_file)

    parser = argparse.ArgumentParser(description="yutto 一个可爱且任性的 B 站视频下载器", prog="yutto")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s {yutto_version}")
    # 如果需要创建其他子命令可参考
    # https://stackoverflow.com/questions/29998417/create-parser-with-subcommands-in-argparse-customize-positional-arguments
    parser.add_argument("url", help="视频主页 url 或 url 列表（需使用 file scheme）")
    group_basic = parser.add_argument_group("basic", "基础参数")
    group_basic.add_argument(
        "-n", "--num-workers", type=int, default=settings.basic.num_workers, help="同时用于下载的最大 Worker 数"
    )
    group_basic.add_argument(
        "-q",
        "--video-quality",
        default=settings.basic.video_quality,
        choices=video_quality_priority_default,
        type=int,
        help="视频清晰度等级（127:8K, 126:Dolby Vision, 125:HDR, 120:4K, 116:1080P60, 112:1080P+, 100:智能修复, 80:1080P, 74:720P60, 64:720P, 32:480P, 16:360P）",
    )
    group_basic.add_argument(
        "-aq",
        "--audio-quality",
        default=settings.basic.audio_quality,
        choices=audio_quality_priority_default,
        type=int,
        help="音频码率等级（30251:Hi-Res, 30255:Dolby Audio, 30250:Dolby Atmos, 30280:320kbps, 30232:128kbps, 30216:64kbps）",
    )
    group_basic.add_argument(
        "--vcodec",
        default=settings.basic.vcodec,
        metavar="DOWNLOAD_VCODEC:SAVE_VCODEC",
        help="视频编码格式（<下载格式>:<生成格式>）",
    )
    group_basic.add_argument(
        "--acodec",
        default=settings.basic.acodec,
        metavar="DOWNLOAD_ACODEC:SAVE_ACODEC",
        help="音频编码格式（<下载格式>:<生成格式>）",
    )
    group_basic.add_argument(
        "--download-vcodec-priority",
        default=settings.basic.download_vcodec_priority,
        type=lambda codecs: codecs.split(",") if codecs != "auto" else None,
        help="视频编码格式优先级，使用 `,` 分隔，如 `hevc,avc,av1`，默认为 `auto`，即根据 vcodec 中「下载编码」自动推断",
    )
    group_basic.add_argument(
        "--output-format",
        default=settings.basic.output_format,
        choices=["infer", "mp4", "mkv", "mov"],
        help="输出格式（infer 为自动推断）",
    )
    group_basic.add_argument(
        "--output-format-audio-only",
        default=settings.basic.output_format_audio_only,
        choices=["infer", "m4a", "aac", "mp3", "flac", "mp4", "mkv", "mov"],
        help="仅包含音频流时所使用的输出格式（infer 为自动推断）",
    )
    group_basic.add_argument(
        "-df",
        "--danmaku-format",
        default=settings.basic.danmaku_format,
        choices=["xml", "ass", "protobuf"],
        help="弹幕类型",
    )
    group_basic.add_argument(
        "-bs",
        "--block-size",
        default=settings.basic.block_size,
        type=float,
        help="分块下载时各块大小，单位为 MiB，默认为 0.5MiB",
    )
    group_basic.add_argument(
        "-w", "--overwrite", default=settings.basic.overwrite, action="store_true", help="强制覆盖已下载内容"
    )
    group_basic.add_argument(
        "-x",
        "--proxy",
        default=settings.basic.proxy,
        help="设置代理（auto 为系统代理、no 为不使用代理、当然也可以设置代理值）",
    )
    group_basic.add_argument(
        "-d",
        "--dir",
        default=path_from_cli(settings.basic.dir),
        type=path_from_cli,
        help="下载目录，默认为运行目录",
    )
    group_basic.add_argument(
        "--tmp-dir",
        default=map_some(path_from_cli, settings.basic.tmp_dir),
        type=path_from_cli,
        help="用来存放下载过程中临时文件的目录，默认为下载目录",
    )
    group_basic.add_argument("-c", "--sessdata", default=settings.basic.sessdata, help="Cookies 中的 SESSDATA 字段")
    group_basic.add_argument(
        "-tp", "--subpath-template", default=settings.basic.subpath_template, help="多级目录的存储路径模板"
    )
    group_basic.add_argument(
        "-af",
        "--alias-file",
        dest="aliases",
        type=alias_parser,
        default=settings.basic.aliases,
        help="设置 url 别名文件路径",
    )
    group_basic.add_argument(
        "--metadata-format-premiered",
        default=settings.basic.metadata_format_premiered,
        help="专用于 metadata 文件中 premiered 字段的日期格式",
    )
    group_basic.add_argument(
        "--download-interval", default=settings.basic.download_interval, type=int, help="设置下载间隔，单位为秒"
    )
    group_basic.add_argument(
        "--banned-mirrors-pattern",
        default=settings.basic.banned_mirrors_pattern,
        help="禁用下载链接的镜像源，使用正则匹配",
    )
    group_basic.add_argument("--no-color", default=settings.basic.no_color, action="store_true", help="不使用颜色")
    group_basic.add_argument(
        "--no-progress", default=settings.basic.no_progress, action="store_true", help="不显示进度条"
    )
    group_basic.add_argument("--debug", default=settings.basic.debug, action="store_true", help="启用 debug 模式")
    group_basic.add_argument(
        "--vip-strict", default=settings.basic.vip_strict, action="store_true", help="启用严格检查大会员生效"
    )
    group_basic.add_argument(
        "--login-strict", default=settings.basic.login_strict, action="store_true", help="启用严格检查登录状态"
    )

    # 资源选择
    group_resource = parser.add_argument_group("resource", "资源选择参数")
    group_resource.add_argument(
        "--video-only",
        dest="require_audio",
        action=create_select_required_action(deselect=["audio"]),
        help="仅下载视频流",
    )
    group_resource.add_argument(
        "--audio-only",
        dest="require_video",
        action=create_select_required_action(deselect=["video"]),
        help="仅下载音频流",
    )  # 视频和音频是反选对方，而不是其余反选所有的
    group_resource.add_argument(
        "--no-danmaku",
        dest="require_danmaku",
        action=create_select_required_action(deselect=["danmaku"]),
        help="不生成弹幕文件",
    )
    group_resource.add_argument(
        "--danmaku-only",
        dest="require_danmaku",
        action=create_select_required_action(select=["danmaku"], deselect=invert_selection(["danmaku"])),
        help="仅生成弹幕文件",
    )
    group_resource.add_argument(
        "--no-subtitle",
        dest="require_subtitle",
        action=create_select_required_action(deselect=["subtitle"]),
        help="不生成字幕文件",
    )
    group_resource.add_argument(
        "--subtitle-only",
        dest="require_subtitle",
        action=create_select_required_action(select=["subtitle"], deselect=invert_selection(["subtitle"])),
        help="仅生成字幕文件",
    )
    group_resource.add_argument(
        "--with-metadata",
        dest="require_metadata",
        action=create_select_required_action(select=["metadata"]),
        help="生成元数据文件",
    )
    group_resource.add_argument(
        "--metadata-only",
        dest="require_metadata",
        action=create_select_required_action(select=["metadata"], deselect=invert_selection(["metadata"])),
        help="仅生成元数据文件",
    )
    group_resource.add_argument(
        "--no-cover",
        dest="require_cover",
        action=create_select_required_action(deselect=["cover"]),
        help="不生成封面",
    )
    group_resource.add_argument(
        "--cover-only",
        dest="require_cover",
        action=create_select_required_action(select=["cover"], deselect=invert_selection(["cover"])),
        help="仅生成封面",
    )
    group_resource.add_argument(
        "--no-chapter-info",
        dest="require_chapter_info",
        action=create_select_required_action(deselect=["chapter_info"]),
        help="不封装章节信息",
    )
    group_resource.add_argument(
        "--save-cover",
        default=settings.resource.save_cover,
        action="store_true",
        help="生成视频流封面后单独保存封面文件",
    )
    group_resource.set_defaults(
        require_video=settings.resource.require_video,
        require_audio=settings.resource.require_audio,
        require_subtitle=settings.resource.require_subtitle,
        require_metadata=settings.resource.require_metadata,
        require_danmaku=settings.resource.require_danmaku,
        require_cover=settings.resource.require_cover,
        require_chapter_info=settings.resource.require_chapter_info,
    )

    # 弹幕设置
    group_danmaku = parser.add_argument_group("danmaku", "弹幕设置参数")
    group_danmaku.add_argument("--danmaku-font-size", type=int, default=settings.danmaku.font_size, help="弹幕字体大小")
    group_danmaku.add_argument("--danmaku-font", default=settings.danmaku.font, help="弹幕字体")
    group_danmaku.add_argument("--danmaku-opacity", type=float, default=settings.danmaku.opacity, help="弹幕不透明度")
    group_danmaku.add_argument(
        "--danmaku-display-region-ratio",
        help="弹幕显示区域与视频高度的比例",
        type=float,
        default=settings.danmaku.display_region_ratio,
    )
    group_danmaku.add_argument("--danmaku-speed", help="弹幕速度", type=float, default=settings.danmaku.speed)
    group_danmaku.add_argument(
        "--danmaku-block-top", action="store_true", default=settings.danmaku.block_top, help="屏蔽顶部弹幕"
    )
    group_danmaku.add_argument(
        "--danmaku-block-bottom", default=settings.danmaku.block_bottom, action="store_true", help="屏蔽底部弹幕"
    )
    group_danmaku.add_argument(
        "--danmaku-block-scroll", default=settings.danmaku.block_scroll, action="store_true", help="屏蔽滚动弹幕"
    )
    group_danmaku.add_argument(
        "--danmaku-block-reverse", default=settings.danmaku.block_reverse, action="store_true", help="屏蔽逆向弹幕"
    )
    group_danmaku.add_argument(
        "--danmaku-block-fixed",
        default=settings.danmaku.block_fixed,
        action="store_true",
        help="屏蔽固定弹幕（顶部、底部）",
    )
    group_danmaku.add_argument(
        "--danmaku-block-special", default=settings.danmaku.block_special, action="store_true", help="屏蔽高级弹幕"
    )
    group_danmaku.add_argument(
        "--danmaku-block-colorful", default=settings.danmaku.block_colorful, action="store_true", help="屏蔽彩色弹幕"
    )
    group_danmaku.add_argument(
        "--danmaku-block-keyword-patterns",
        default=settings.danmaku.block_keyword_patterns,
        type=lambda patterns: [pattern.strip() for pattern in patterns.split(",")],
        help="屏蔽匹配关键词的弹幕，使用逗号分隔",
    )

    # 仅批量下载使用
    group_batch = parser.add_argument_group("batch", "批量下载参数")
    group_batch.add_argument("-b", "--batch", action="store_true", help="批量下载")
    group_batch.add_argument("-p", "--episodes", default="1~-1", help="选集")
    group_batch.add_argument(
        "-s",
        "--with-section",
        action="store_true",
        default=settings.batch.with_section,
        help="同时下载附加剧集（PV、预告以及特别篇等专区内容）",
    )
    group_batch.add_argument(
        "--batch-filter-start-time",
        default=settings.batch.batch_filter_start_time,
        help="只下载该时间之后（包含临界值）发布的稿件",
    )
    group_batch.add_argument(
        "--batch-filter-end-time",
        default=settings.batch.batch_filter_end_time,
        help="只下载该时间之前（不包含临界值）发布的稿件",
    )

    # 仅任务列表中使用
    group_batch_file = parser.add_argument_group("batch file", "批量下载文件参数")
    group_batch_file.add_argument("--no-inherit", action="store_true", help="不继承父级参数")

    # 配置路径（占位用的，config 已经在 pre parser 里解析过了）
    group_config = parser.add_argument_group("config", "配置文件参数")
    group_config.add_argument("--config", help="配置文件路径")

    return parser


def create_select_required_action(
    select: list[DownloadResourceType] | None = None, deselect: list[DownloadResourceType] | None = None
):
    selected_items = select or []
    deselected_items = deselect or []

    class SelectRequiredAction(argparse.Action):
        def __init__(self, option_strings: str, dest: str, nargs: int | str | None = None, **kwargs: Any):
            if nargs is not None:
                raise ValueError("nargs not allowed")
            super().__init__(option_strings, dest, nargs=0, **kwargs)

        def __call__(
            self,
            parser: argparse.ArgumentParser,
            namespace: argparse.Namespace,
            values: str | Sequence[str] | None,
            option_string: str | None = None,
        ):
            for select_item in selected_items:
                setattr(namespace, f"require_{select_item}", True)
            for deselect_item in deselected_items:
                setattr(namespace, f"require_{deselect_item}", False)

    return SelectRequiredAction


def invert_selection(select: list[DownloadResourceType]) -> list[DownloadResourceType]:
    return [tp for tp in DOWNLOAD_RESOURCE_TYPES if tp not in select]



from __future__ import annotations

import asyncio
import copy
import os
import re
import shlex
import sys
from typing import TYPE_CHECKING, Callable

import httpx
from biliass import BlockOptions

from yutto.cli.cli import cli
from yutto.exceptions import ErrorCode
from yutto.extractor import (
    BangumiBatchExtractor,
    BangumiExtractor,
    CheeseBatchExtractor,
    CheeseExtractor,
    CollectionExtractor,
    FavouritesExtractor,
    SeriesExtractor,
    UgcVideoBatchExtractor,
    UgcVideoExtractor,
    UserAllFavouritesExtractor,
    UserAllUgcVideosExtractor,
    UserWatchLaterExtractor,
)
from yutto.processor.downloader import DownloadState, start_downloader
from yutto.processor.parser import file_scheme_parser
from yutto.processor.path_resolver import create_unique_path_resolver
from yutto.utils.asynclib import sleep_with_status_bar_refresh
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.danmaku import DanmakuOptions
from yutto.utils.fetcher import Fetcher, FetcherContext, create_client
from yutto.utils.funcutils import as_sync
from yutto.utils.time import TIME_FULL_FMT
from yutto.validator import (
    initial_validation,
    validate_basic_arguments,
    validate_batch_arguments,
    validate_user_info,
)

if TYPE_CHECKING:
    import argparse

    from yutto._typing import EpisodeData


def main():
    parser = cli()
    args = parser.parse_args()
    ctx = FetcherContext()
    initial_validation(ctx, args)
    args_list = flatten_args(args, parser)
    try:
        run(ctx, args_list)
    except (SystemExit, KeyboardInterrupt, asyncio.exceptions.CancelledError):
        Logger.info("已终止下载，再次运行即可继续下载～")
        sys.exit(ErrorCode.PAUSED_DOWNLOAD.value)


@as_sync
async def run(ctx: FetcherContext, args_list: list[argparse.Namespace]):
    ctx.set_fetch_semaphore(fetch_workers=8)
    unique_path = create_unique_path_resolver()
    async with create_client(
        cookies=ctx.cookies,
        trust_env=ctx.trust_env,
        proxy=ctx.proxy,
    ) as client:
        if len(args_list) > 1:
            Logger.info(f"列表里共检测到 {len(args_list)} 项")

        for i, args in enumerate(args_list):
            if len(args_list) > 1:
                Logger.custom(f"列表项 {args.url}", Badge(f"[{i + 1}/{len(args_list)}]", fore="black", back="cyan"))

            # 验证批量参数
            if args.batch:
                validate_batch_arguments(args)

            # 初始化各种提取器
            extractors = (
                [
                    UgcVideoBatchExtractor(),  # 投稿全集
                    BangumiBatchExtractor(),  # 番剧全集
                    CheeseBatchExtractor(),  # 课程全集
                    FavouritesExtractor(),  # 用户单一收藏
                    UserAllFavouritesExtractor(),  # 用户全部收藏
                    SeriesExtractor(),  # 视频列表
                    CollectionExtractor(),  # 视频合集
                    UserAllUgcVideosExtractor(),  # 个人空间，由于个人空间的正则包含了收藏夹，所以需要放在收藏夹之后
                    UserWatchLaterExtractor(),  # 用户稍后再看
                ]
                if args.batch
                else [
                    UgcVideoExtractor(),  # 投稿单集
                    BangumiExtractor(),  # 番剧单话
                    CheeseExtractor(),  # 课程单集
                ]
            )
            url: str = args.url
            # 将 shortcut 转为完整 url
            for extractor in extractors:
                matched, url = extractor.resolve_shortcut(url)
                if matched:
                    break

            # 在开始前校验，减少对第一个视频的请求
            if not await validate_user_info(ctx, {"is_login": args.login_strict, "vip_status": args.vip_strict}):
                Logger.error("启用了严格校验大会员或登录模式，请检查 SESSDATA 或大会员状态！")
                sys.exit(ErrorCode.NOT_LOGIN_ERROR.value)
            # 重定向到可识别的 url
            try:
                url = await Fetcher.get_redirected_url(ctx, client, url)
            except httpx.InvalidURL:
                Logger.error(f"无效的 url({url})～请检查一下链接是否正确～")
                sys.exit(ErrorCode.WRONG_URL_ERROR.value)
            except httpx.UnsupportedProtocol:
                error_text = f"无效的 url 协议（{url}）～请检查一下链接协议是否正确"
                if not args.batch:
                    error_text += (
                        "，如使用裸 id 功能，请确认该类型 id 是否支持当前单话模式，如不支持需要添加 `-b` 以使用批量模式"
                    )
                Logger.error(error_text)
                sys.exit(ErrorCode.WRONG_URL_ERROR.value)

            # 提取信息，构造解析任务～
            for extractor in extractors:
                if extractor.match(url):
                    download_list = await extractor(ctx, client, args)
                    break
            else:
                if args.batch:
                    # TODO: 指向文档中受支持的列表部分
                    Logger.error("url 不正确呦～")
                else:
                    Logger.error("url 不正确，也许该 url 仅支持批量下载，如果是这样，请使用参数 -b～")
                sys.exit(ErrorCode.WRONG_URL_ERROR.value)

            current_download_state = DownloadState.SKIP

            # 下载～
            for i, episode_data_coro in enumerate(download_list):
                if episode_data_coro is None:
                    continue

                # 中途校验，因为批量下载时可能会失效
                if not await validate_user_info(ctx, {"is_login": args.login_strict, "vip_status": args.vip_strict}):
                    Logger.error("启用了严格校验大会员或登录模式，请检查 SESSDATA 或大会员状态！")
                    sys.exit(ErrorCode.NOT_LOGIN_ERROR.value)

                if current_download_state != DownloadState.SKIP and args.download_interval > 0:
                    Logger.info(f"下载间隔 {args.download_interval} 秒")
                    await sleep_with_status_bar_refresh(args.download_interval)

                # 这时候才真正开始解析链接
                episode_data = await episode_data_coro
                if episode_data is None:
                    continue
                # 保证路径唯一
                episode_data = ensure_unique_path(episode_data, unique_path)
                if args.batch:
                    Logger.custom(
                        f"{episode_data['filename']}",
                        Badge(f"[{i + 1}/{len(download_list)}]", fore="black", back="cyan"),
                    )

                current_download_state = await start_downloader(
                    ctx,
                    client,
                    episode_data,
                    {
                        "require_video": args.require_video,
                        "require_chapter_info": args.require_chapter_info,
                        "video_quality": args.video_quality,
                        "video_download_codec": args.vcodec.split(":")[0],
                        "video_save_codec": args.vcodec.split(":")[1],
                        "video_download_codec_priority": args.download_vcodec_priority,
                        "require_audio": args.require_audio,
                        "audio_quality": args.audio_quality,
                        "audio_download_codec": args.acodec.split(":")[0],
                        "audio_save_codec": args.acodec.split(":")[1],
                        "output_format": args.output_format,
                        "output_format_audio_only": args.output_format_audio_only,
                        "overwrite": args.overwrite,
                        "block_size": int(args.block_size * 1024 * 1024),
                        "num_workers": args.num_workers,
                        "save_cover": args.save_cover,
                        "metadata_format": {
                            "premiered": args.metadata_format_premiered,
                            "dateadded": TIME_FULL_FMT,
                        },
                        "banned_mirrors_pattern": args.banned_mirrors_pattern,
                        "danmaku_options": parse_danmaku_options(args),
                    },
                )
                Logger.new_line()
            Logger.new_line()


def flatten_args(args: argparse.Namespace, parser: argparse.ArgumentParser) -> list[argparse.Namespace]:
    """递归展平列表参数"""
    args = copy.copy(args)
    validate_basic_arguments(args)
    # 查看是否存在于 alias 中
    alias_map: dict[str, str] = args.aliases if args.aliases is not None else {}
    if args.url in alias_map:
        args.url = alias_map[args.url]

    # 是否为下载列表
    if re.match(r"file://", args.url) or os.path.isfile(args.url):  # noqa: PTH113
        args_list: list[argparse.Namespace] = []
        # TODO: 如果是相对路径，需要相对于当前 list 路径
        for line in file_scheme_parser(args.url):
            local_args = parser.parse_args(shlex.split(line), args)
            if local_args.no_inherit:
                local_args = parser.parse_args(shlex.split(line))
            Logger.debug(f"列表参数: {local_args}")
            args_list += flatten_args(local_args, parser)
        return args_list
    else:
        return [args]


def ensure_unique_path(episode_data: EpisodeData, unique_name_resolver: Callable[[str], str]) -> EpisodeData:
    original_filename = episode_data["filename"]
    new_name = unique_name_resolver(original_filename)
    episode_data["filename"] = new_name
    if original_filename != new_name:
        Logger.warning(f"文件名重复，已重命名为 {new_name}")
    return episode_data


def parse_danmaku_options(args: argparse.Namespace) -> DanmakuOptions:
    block_options = BlockOptions(
        block_top=args.danmaku_block_top or args.danmaku_block_fixed,
        block_bottom=args.danmaku_block_bottom or args.danmaku_block_fixed,
        block_scroll=args.danmaku_block_scroll,
        block_reverse=args.danmaku_block_reverse,
        block_special=args.danmaku_block_special,
        block_colorful=args.danmaku_block_colorful,
        block_keyword_patterns=(args.danmaku_block_keyword_patterns if args.danmaku_block_keyword_patterns else []),
    )
    return DanmakuOptions(
        font_size=args.danmaku_font_size,
        font=args.danmaku_font,
        opacity=args.danmaku_opacity,
        display_region_ratio=args.danmaku_display_region_ratio,
        speed=args.danmaku_speed,
        block_options=block_options,
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
import urllib
import urllib.request
from pathlib import Path

from yutto.utils.console.logger import Logger


def path_from_cli(path: str) -> Path:
    """从命令行参数获取路径，支持 ~，以便配置中使用 ~"""
    return Path(path).expanduser()


def is_comment(line: str) -> bool:
    """判断文件某行是否为注释"""
    if line.startswith("#"):
        return True
    return False


def alias_parser(file_path: str) -> dict[str, str]:
    result: dict[str, str] = {}
    re_alias_splitter = re.compile(r"[\s=]")
    with path_from_cli(file_path).open("r") as f_alias:
        for line in f_alias:
            line = line.strip()
            if not line or is_comment(line):
                continue
            alias, url = re_alias_splitter.split(line, maxsplit=1)
            result[alias] = url
    return result


def file_scheme_parser(url: str) -> list[str]:
    file_url: str = urllib.parse.urlparse(url).path  # type: ignore
    file_path = path_from_cli(urllib.request.url2pathname(file_url))  # type: ignore
    Logger.info(f"解析下载列表 {file_path} 中...")
    result: list[str] = []
    with file_path.open("r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or is_comment(line):
                continue
            result.append(line)
    return result



from __future__ import annotations

import pytest

from yutto._typing import AId, BvId, CId, EpisodeId
from yutto.api.ugc_video import (
    get_ugc_video_info,
    get_ugc_video_list,
    get_ugc_video_playurl,
    get_ugc_video_subtitles,
)
from yutto.utils.fetcher import FetcherContext, create_client
from yutto.utils.funcutils import as_sync


@pytest.mark.api
@pytest.mark.ci_skip
@as_sync
async def test_get_ugc_video_info():
    bvid = BvId("BV1q7411v7Vd")
    aid = AId("84271171")
    avid = bvid
    episode_id = EpisodeId("300998")
    ctx = FetcherContext()
    async with create_client() as client:
        video_info = await get_ugc_video_info(ctx, client, avid=avid)
        assert video_info["avid"] == aid or video_info["avid"] == bvid
        assert video_info["aid"] == aid
        assert video_info["bvid"] == bvid
        assert video_info["episode_id"] == episode_id
        assert video_info["is_bangumi"] is True
        assert video_info["cid"] == CId("144541892")
        assert video_info["title"] == "【独播】我的三体之章北海传 第1集"


@pytest.mark.api
@as_sync
async def test_get_ugc_video_title():
    avid = BvId("BV1vZ4y1M7mQ")
    ctx = FetcherContext()
    async with create_client() as client:
        title = (await get_ugc_video_list(ctx, client, avid))["title"]
        assert title == "用 bilili 下载 B 站视频"


@pytest.mark.api
@as_sync
async def test_get_ugc_video_list():
    avid = BvId("BV1vZ4y1M7mQ")
    ctx = FetcherContext()
    async with create_client() as client:
        ugc_video_list = (await get_ugc_video_list(ctx, client, avid))["pages"]
        assert ugc_video_list[0]["id"] == 1
        assert ugc_video_list[0]["name"] == "bilili 特性以及使用方法简单介绍"
        assert ugc_video_list[0]["cid"] == CId("222190584")
        assert ugc_video_list[0]["metadata"] is not None
        assert ugc_video_list[0]["metadata"]["title"] == "bilili 特性以及使用方法简单介绍"
        assert ugc_video_list[0]["metadata"]["website"] == "https://www.bilibili.com/video/BV1vZ4y1M7mQ"

        assert ugc_video_list[1]["id"] == 2
        assert ugc_video_list[1]["name"] == "bilili 环境配置方法"
        assert ugc_video_list[1]["cid"] == CId("222200470")
        assert ugc_video_list[1]["metadata"] is not None
        assert ugc_video_list[1]["metadata"]["title"] == "bilili 环境配置方法"
        assert ugc_video_list[0]["metadata"]["website"] == "https://www.bilibili.com/video/BV1vZ4y1M7mQ"


@pytest.mark.api
@pytest.mark.ci_skip
@as_sync
async def test_get_ugc_video_playurl():
    avid = BvId("BV1vZ4y1M7mQ")
    cid = CId("222190584")
    ctx = FetcherContext()
    async with create_client() as client:
        playlist = await get_ugc_video_playurl(ctx, client, avid, cid)
        assert len(playlist[0]) > 0
        assert len(playlist[1]) > 0


# The latest subtitle API needs login, so this test is skipped.
# We need to find a way to test theses APIs.
@pytest.mark.skip
@pytest.mark.api
@as_sync
async def test_get_ugc_video_subtitles():
    avid = BvId("BV1Ra411A7kN")
    cid = CId("253246252")
    ctx = FetcherContext()
    async with create_client() as client:
        subtitles = await get_ugc_video_subtitles(ctx, client, avid=avid, cid=cid)
        assert len(subtitles) > 0
        assert len(subtitles[0]["lines"]) > 0



from __future__ import annotations

import pytest

from yutto._typing import AId, BvId, FId, MId, SeriesId
from yutto.api.space import (
    get_all_favourites,
    get_favourite_avids,
    get_favourite_info,
    get_medialist_avids,
    get_medialist_title,
    get_user_name,
    get_user_space_all_videos_avids,
)
from yutto.utils.fetcher import FetcherContext, create_client
from yutto.utils.funcutils import as_sync


@pytest.mark.api
@pytest.mark.ignore
@as_sync
async def test_get_user_space_all_videos_avids():
    mid = MId("100969474")
    ctx = FetcherContext()
    async with create_client() as client:
        all_avid = await get_user_space_all_videos_avids(ctx, client, mid=mid)
        assert len(all_avid) > 0
        assert AId("371660125") in all_avid or BvId("BV1vZ4y1M7mQ") in all_avid


@pytest.mark.api
@pytest.mark.ignore
@as_sync
async def test_get_user_name():
    mid = MId("100969474")
    ctx = FetcherContext()
    async with create_client() as client:
        username = await get_user_name(ctx, client, mid=mid)
        assert username == "时雨千陌"


@pytest.mark.api
@as_sync
async def test_get_favourite_info():
    fid = FId("1306978874")
    ctx = FetcherContext()
    async with create_client() as client:
        fav_info = await get_favourite_info(ctx, client, fid=fid)
        assert fav_info["fid"] == fid
        assert fav_info["title"] == "Test"


@pytest.mark.api
@as_sync
async def test_get_favourite_avids():
    fid = FId("1306978874")
    ctx = FetcherContext()
    async with create_client() as client:
        avids = await get_favourite_avids(ctx, client, fid=fid)
        assert AId("456782499") in avids or BvId("BV1o541187Wh") in avids


@pytest.mark.api
@as_sync
async def test_all_favourites():
    mid = MId("100969474")
    ctx = FetcherContext()
    async with create_client() as client:
        fav_list = await get_all_favourites(ctx, client, mid=mid)
        assert {"fid": FId("1306978874"), "title": "Test"} in fav_list


@pytest.mark.api
@as_sync
async def test_get_medialist_avids():
    series_id = SeriesId("1947439")
    mid = MId("100969474")
    ctx = FetcherContext()
    async with create_client() as client:
        avids = await get_medialist_avids(ctx, client, series_id=series_id, mid=mid)
        assert avids == [BvId("BV1Y441167U2"), BvId("BV1vZ4y1M7mQ")]


@pytest.mark.api
@as_sync
async def test_get_medialist_title():
    series_id = SeriesId("1947439")
    ctx = FetcherContext()
    async with create_client() as client:
        title = await get_medialist_title(ctx, client, series_id=series_id)
        assert title == "一个小视频列表～"
